@@ -113,7 +113,7 @@ static char sccsid[] = "@(#)syslogd.c	5.27 (Berkeley) 10/10/88";
  *	priority information is encoded as a leader on the messages this
  *	was causing lines to be placed in erroneous files.
  *
- *	Also added a patch from Shane Alderton (shane@scs.apana.org.au) to
+ *	Also added a patch from Shane Alderton (shane@ion.apana.org.au) to
  *	correct a problem with syslogd dumping core when an attempt was made
  *	to write log messages to a logged-on user.  Thank you.
  *
@@ -351,6 +351,66 @@ static char sccsid[] = "@(#)syslogd.c	5.27 (Berkeley) 10/10/88";
  * Wed Feb 25 13:21:44 CET 1998: Martin Schulze <joey@infodrom.north.de>
  *	Corrected Topi's patch as it prevented forwarding during
  *	startup due to an unknown LogPort.
+ *
+ * Sat Oct 10 20:01:48 CEST 1998: Martin Schulze <joey@infodrom.north.de>
+ *	Added support for TESTING define which will turn syslogd into
+ *	stdio-mode used for debugging.
+ *
+ * Sun Oct 11 20:16:59 CEST 1998: Martin Schulze <joey@infodrom.north.de>
+ *	Reworked the initialization/fork code.  Now the parent
+ *	process activates a signal handler which the daughter process
+ *	will raise if it is initialized.  Only after that one the
+ *	parent process may exit.  Otherwise klogd might try to flush
+ *	its log cache while syslogd can't receive the messages yet.
+ *
+ * Mon Oct 12 13:30:35 CEST 1998: Martin Schulze <joey@infodrom.north.de>
+ *	Redirected some error output with regard to argument parsing to
+ *	stderr.
+ *
+ * Mon Oct 12 14:02:51 CEST 1998: Martin Schulze <joey@infodrom.north.de>
+ *	Applied patch provided vom Topi Miettinen with regard to the
+ *	people from OpenBSD.  This provides the additional '-a'
+ *	argument used for specifying additional UNIX domain sockets to
+ *	listen to.  This is been used with chroot()'ed named's for
+ *	example.  See for http://www.psionic.com/papers/dns.html
+ *
+ * Mon Oct 12 18:29:44 CEST 1998: Martin Schulze <joey@infodrom.north.de>
+ *	Added `ftp' facility which was introduced in glibc version 2.
+ *	It's #ifdef'ed so won't harm with older libraries.
+ *
+ * Mon Oct 12 19:59:21 MET DST 1998: Martin Schulze <joey@infodrom.north.de>
+ *	Code cleanups with regard to bsd -> posix transition and
+ *	stronger security (buffer length checking).  Thanks to Topi
+ *	Miettinen <tom@medialab.sonera.net>
+ *	. index() --> strchr()
+ *	. sprintf() --> snprintf()
+ *	. bcopy() --> memcpy()
+ *	. bzero() --> memset()
+ *	. UNAMESZ --> UT_NAMESIZE
+ *	. sys_errlist --> strerror()
+ *
+ * Mon Oct 12 20:22:59 CEST 1998: Martin Schulze <joey@infodrom.north.de>
+ *	Added support for setutent()/getutent()/endutend() instead of
+ *	binary reading the UTMP file.  This is the the most portable
+ *	way.  This allows /var/run/utmp format to change, even to a
+ *	real database or utmp daemon. Also if utmp file locking is
+ *	implemented in libc, syslog will use it immediately.  Thanks
+ *	to Topi Miettinen <tom@medialab.sonera.net>.
+ *
+ * Mon Oct 12 20:49:18 MET DST 1998: Martin Schulze <joey@infodrom.north.de>
+ *	Avoid logging of SIGCHLD when syslogd is in the process of
+ *	exiting and closing its files.  Again thanks to Topi.
+ *
+ * Mon Oct 12 22:18:34 CEST 1998: Martin Schulze <joey@infodrom.north.de>
+ *	Modified printline() to support 8bit characters - such as
+ *	russion letters.  Thanks to Vladas Lapinskas <lapinskas@mail.iae.lt>.
+ *
+ * Sat Nov 14 02:29:37 CET 1998: Martin Schulze <joey@infodrom.north.de>
+ *	``-m 0'' now turns of MARK logging entirely.
+ *
+ * Tue Jan 19 01:04:18 MET 1999: Martin Schulze <joey@infodrom.north.de>
+ *	Finally fixed an error with `-a' processing, thanks to Topi
+ *	Miettinen <tom@medialab.sonera.net>.
  */
 
 
@@ -373,6 +433,7 @@ static char sccsid[] = "@(#)syslogd.c	5.27 (Berkeley) 10/10/88";
 #include <string.h>
 #include <setjmp.h>
 #include <stdarg.h>
+#include <time.h>
 
 #define SYSLOG_NAMES
 #include <sys/syslog.h>
@@ -400,7 +461,9 @@ static char sccsid[] = "@(#)syslogd.c	5.27 (Berkeley) 10/10/88";
 #include <arpa/nameser.h>
 #include <arpa/inet.h>
 #include <resolv.h>
+#ifndef TESTING
 #include "pidfile.h"
+#endif
 #include "version.h"
 
 #if defined(__linux__)
@@ -456,19 +519,28 @@ static char sccsid[] = "@(#)syslogd.c	5.27 (Berkeley) 10/10/88";
 #define _PATH_LOG	"/dev/log"
 #endif
 
-char	*LogName = _PATH_LOG;
 char	*ConfFile = _PATH_LOGCONF;
 char	*PidFile = _PATH_LOGPID;
 char	ctty[] = _PATH_CONSOLE;
 
 char	**parts;
 
-int inetm = 0, funix = -1;
+int inetm = 0;
 static int debugging_on = 0;
 static int nlogs = -1;
 static int restart = 0;
 
-#define UNAMESZ		8	/* length of a login name */
+#define MAXFUNIX	20
+
+int nfunix = 1;
+char *funixn[MAXFUNIX] = { _PATH_LOG };
+int funix[MAXFUNIX] = { -1, };
+
+#ifdef UT_NAMESIZE
+# define UNAMESZ	UT_NAMESIZE	/* length of a login name */
+#else
+# define UNAMESZ	8	/* length of a login name */
+#endif
 #define MAXUNAMES	20	/* maximum number of user names */
 #define MAXFNAME	200	/* max file pathname length */
 
@@ -604,6 +676,9 @@ struct code	FacNames[] = {
 	{"syslog",       LOG_SYSLOG},
 	{"user",         LOG_USER},
 	{"uucp",         LOG_UUCP},
+#if defined(LOG_FTP)
+	{"ftp",          LOG_FTP},
+#endif
 	{"local0",       LOG_LOCAL0},
 	{"local1",       LOG_LOCAL1},
 	{"local2",       LOG_LOCAL2},
@@ -631,11 +706,7 @@ char	**LocalHosts = NULL;	/* these hosts are logged with their hostname */
 int	NoHops = 1;		/* Can we bounce syslog messages through an
 				   intermediate host. */
 
-extern	int errno, sys_nerr;
-#if !defined(__GLIBC__)
-extern	char *sys_errlist[];
-#endif /* __GLIBC__ */
-extern	char *ctime(), *index();
+extern	int errno;
 
 /* Function prototypes. */
 int main(int argc, char **argv);
@@ -655,6 +726,9 @@ void domark();
 void debug_switch();
 void logerror(char *type);
 void die(int sig);
+#ifndef TESTING
+void doexit(int sig);
+#endif
 void init();
 void cfline(char *line, register struct filed *f);
 int decode(char *name, struct code *codetab);
@@ -666,7 +740,7 @@ static void allocate_log(void);
 void sighup_handler();
 
 #ifdef SYSLOG_UNIXAF
-static int create_unix_socket();
+static int create_unix_socket(const char *path);
 #endif
 #ifdef SYSLOG_INET
 static int create_inet_socket();
@@ -681,11 +755,22 @@ int main(argc, argv)
 #if !defined(__GLIBC__)
 	int len, num_fds;
 #else /* __GLIBC__ */
+#ifndef TESTING
 	size_t len;
+#endif
 	int num_fds;
 #endif /* __GLIBC__ */
+	/*
+	 * It took me quite some time to figure out how this is
+	 * supposed to work so I guess I should better write it down.
+	 * unixm is a list of file descriptors from which one can
+	 * read().  This is in contrary to readfds which is a list of
+	 * file descriptors where activity is monitored by select()
+	 * and from which one cannot read().  -Joey
+	 */
 	fd_set unixm, readfds;
 
+#ifndef TESTING
 	int	fd;
 #ifdef SYSLOG_UNIXAF
 	struct sockaddr_un fromunix;
@@ -693,6 +778,8 @@ int main(argc, argv)
 #ifdef  SYSLOG_INET
 	struct sockaddr_in frominet;
 	char *from;
+#endif
+	pid_t ppid = getpid();
 #endif
 	int ch;
 	struct hostent *hent;
@@ -700,10 +787,24 @@ int main(argc, argv)
 	char line[MAXLINE +1];
 	extern int optind;
 	extern char *optarg;
+	int maxfds;
 
+#ifndef TESTING
 	chdir ("/");
-	while ((ch = getopt(argc, argv, "dhf:l:m:np:rs:v")) != EOF)
+#endif
+	for (i = 1; i < MAXFUNIX; i++) {
+		funixn[i] = "";
+		funix[i]  = -1;
+	}
+
+	while ((ch = getopt(argc, argv, "a:dhf:l:m:np:rs:v")) != EOF)
 		switch((char)ch) {
+		case 'a':
+			if (nfunix < MAXFUNIX)
+				funixn[nfunix++] = optarg;
+			else
+				fprintf(stderr, "Out of descriptors, ignoring %s\n", optarg);
+			break;
 		case 'd':		/* debug */
 			Debug = 1;
 			break;
@@ -715,7 +816,7 @@ int main(argc, argv)
 			break;
 		case 'l':
 			if (LocalHosts) {
-				printf ("Only one -l argument allowed," \
+				fprintf (stderr, "Only one -l argument allowed," \
 					"the first one is taken.\n");
 				break;
 			}
@@ -727,15 +828,15 @@ int main(argc, argv)
 		case 'n':		/* don't fork */
 			NoFork = 1;
 			break;
-		case 'p':		/* path */
-			LogName = optarg;
+		case 'p':		/* path to regular log socket */
+			funixn[0] = optarg;
 			break;
 		case 'r':		/* accept remote messages */
 			AcceptRemote = 1;
 			break;
 		case 's':
 			if (StripDomains) {
-				printf ("Only one -s argument allowed," \
+				fprintf (stderr, "Only one -s argument allowed," \
 					"the first one is taken.\n");
 				break;
 			}
@@ -748,19 +849,31 @@ int main(argc, argv)
 		default:
 			usage();
 		}
-	if (argc -= optind)
+	if ((argc -= optind))
 		usage();
 
+#ifndef TESTING
 	if ( !(Debug || NoFork) )
 	{
 		dprintf("Checking pidfile.\n");
 		if (!check_pid(PidFile))
 		{
-			if (fork())
+			if (fork()) {
 				/*
 				 * Parent process
 				 */
-				exit(0);
+				signal (SIGTERM, doexit);
+				sleep(300);
+				/*
+				 * Not reached unless something major went wrong.  5
+				 * minutes should be a fair amount of time to wait.
+				 * Please note that this procedure is important since
+				 * the father must not exit before syslogd isn't
+				 * initialized or the klogd won't be able to flush its
+				 * logs.  -Joey
+				 */
+				exit(1);
+			}
 			num_fds = getdtablesize();
 			for (i= 0; i < num_fds; i++)
 				(void) close(i);
@@ -773,12 +886,14 @@ int main(argc, argv)
 		}
 	}
 	else
+#endif
 		debugging_on = 1;
 #ifndef SYSV
 	else
 		setlinebuf(stdout);
 #endif
 
+#ifndef TESTING
 	/* tuck my process id away */
 	if ( !Debug )
 	{
@@ -797,11 +912,12 @@ int main(argc, argv)
 			exit(1);
 		}
 	} /* if ( !Debug ) */
+#endif
 
 	consfile.f_type = F_CONSOLE;
 	(void) strcpy(consfile.f_un.f_fname, ctty);
 	(void) gethostname(LocalHostName, sizeof(LocalHostName));
-	if ( (p = index(LocalHostName, '.')) ) {
+	if ( (p = strchr(LocalHostName, '.')) ) {
 		*p++ = '\0';
 		LocalDomain = p;
 	}
@@ -822,9 +938,9 @@ int main(argc, argv)
 		 */
 		hent = gethostbyname(LocalHostName);
 		if ( hent )
-			sprintf(LocalHostName, "%s", hent->h_name);
+			snprintf(LocalHostName, sizeof(LocalHostName), "%s", hent->h_name);
 			
-		if ( (p = index(LocalHostName, '.')) )
+		if ( (p = strchr(LocalHostName, '.')) )
 		{
 			*p++ = '\0';
 			LocalDomain = p;
@@ -845,7 +961,6 @@ int main(argc, argv)
 	(void) signal(SIGALRM, domark);
 	(void) signal(SIGUSR1, Debug ? debug_switch : SIG_IGN);
 	(void) alarm(TIMERINTVL);
-	(void) unlink(LogName);
 
 	/* Create a partial message table for all file descriptors. */
 	num_fds = getdtablesize();
@@ -861,48 +976,76 @@ int main(argc, argv)
 
 	dprintf("Starting.\n");
 	init();
+#ifndef TESTING
 	if ( Debug )
 	{
 		dprintf("Debugging disabled, SIGUSR1 to turn on debugging.\n");
 		debugging_on = 0;
 	}
+	/*
+	 * Send a signal to the parent to it can terminate.
+	 */
+	if (getpid() != ppid)
+		kill (ppid, SIGTERM);
+#endif
 
 	/* Main loop begins here. */
 	FD_ZERO(&unixm);
-	FD_ZERO(&readfds);
 	for (;;) {
 		int nfds;
 		errno = 0;
+		FD_ZERO(&readfds);
+		maxfds = 0;
 #ifdef SYSLOG_UNIXAF
+#ifndef TESTING
 		/*
-		 * Add the Unix Domain Socket to the list of read
+		 * Add the Unix Domain Sockets to the list of read
 		 * descriptors.
 		 */
-		FD_SET(funix, &readfds);
+		/* Copy master connections */
+		for (i = 0; i < nfunix; i++) {
+			if (funix[i] != -1) {
+				FD_SET(funix[i], &readfds);
+				if (funix[i]>maxfds) maxfds=funix[i];
+			}
+		}
+		/* Copy accepted connections */
 		for (nfds= 0; nfds < FD_SETSIZE; ++nfds)
-			if ( FD_ISSET(nfds, &unixm) )
+			if (FD_ISSET(nfds, &unixm)) {
 				FD_SET(nfds, &readfds);
+				if (nfds>maxfds) maxfds=nfds;
+			}
+#endif
 #endif
 #ifdef SYSLOG_INET
+#ifndef TESTING
 		/*
 		 * Add the Internet Domain Socket to the list of read
 		 * descriptors.
 		 */
 		if ( InetInuse && AcceptRemote ) {
 			FD_SET(inetm, &readfds);
+			if (inetm>maxfds) maxfds=inetm;
 			dprintf("Listening on syslog UDP port.\n");
 		}
 #endif
+#endif
+#ifdef TESTING
+		FD_SET(fileno(stdin), &readfds);
+		if (fileno(stdin) > maxfds) maxfds = fileno(stdin);
+
+		dprintf("Listening on stdin.  Press Ctrl-C to interrupt.\n");
+#endif
 
 		if ( debugging_on )
 		{
-			dprintf("Calling select, active file descriptors: ");
-			for (nfds= 0; nfds < FD_SETSIZE; ++nfds)
+			dprintf("Calling select, active file descriptors (max %d): ", maxfds);
+			for (nfds= 0; nfds <= maxfds; ++nfds)
 				if ( FD_ISSET(nfds, &readfds) )
 					dprintf("%d ", nfds);
 			dprintf("\n");
 		}
-		nfds = select(FD_SETSIZE, (fd_set *) &readfds, (fd_set *) NULL,
+		nfds = select(maxfds+1, (fd_set *) &readfds, (fd_set *) NULL,
 				  (fd_set *) NULL, (struct timeval *) NULL);
 		if ( restart )
 		{
@@ -926,22 +1069,23 @@ int main(argc, argv)
 		{
 			dprintf("\nSuccessful select, descriptor count = %d, " \
 				"Activity on: ", nfds);
-			for (nfds= 0; nfds < FD_SETSIZE; ++nfds)
+			for (nfds= 0; nfds <= maxfds; ++nfds)
 				if ( FD_ISSET(nfds, &readfds) )
 					dprintf("%d ", nfds);
 			dprintf(("\n"));
 		}
 
+#ifndef TESTING
 #ifdef SYSLOG_UNIXAF
 		if ( debugging_on )
 		{
 			dprintf("Checking UNIX connections, active: ");
-			for (nfds= 0; nfds < FD_SETSIZE; ++nfds)
+			for (nfds= 0; nfds < maxfds; ++nfds)
 				if ( FD_ISSET(nfds, &unixm) )
 					dprintf("%d ", nfds);
 			dprintf("\n");
 		}
-		for (fd= 0; fd < FD_SETSIZE; ++fd)
+		for (fd= 0; fd <= maxfds; ++fd)
 		  if ( FD_ISSET(fd, &readfds) && FD_ISSET(fd, &unixm) ) {
 			dprintf("Message from UNIX socket #%d.\n", fd);
 			memset(line, '\0', sizeof(line));
@@ -962,27 +1106,30 @@ int main(argc, argv)
 						     strlen(parts[fd]) + 1, \
 						     fd);
 				}
+				/* reset it */
+				for (i = 1; i < nfunix; i++) {
+					if (funix[i] == fd)
+						funix[i] = -1;
+				}
 		    		close(fd);
 		    		FD_CLR(fd, &unixm);
-		    		FD_CLR(fd, &readfds);
 		  	}
 	      	}
 		/* Accept a new unix connection */
-		if (FD_ISSET(funix, &readfds)) {
-			len = sizeof(fromunix);
-			if ((fd = accept(funix, (struct sockaddr *) &fromunix,\
-					 &len)) >= 0) {
-			  	FD_SET(fd, &unixm);
-				dprintf("New UNIX connect assigned to fd: " \
-					"%d.\n", fd);
-				FD_SET(fd, &readfds);
+		for (i = 0; i < nfunix; i++)
+			if (funix[i] != -1 && FD_ISSET(funix[i], &readfds)) {
+				len = sizeof(fromunix);
+				if ((fd = accept(funix[i], (struct sockaddr *) &fromunix,
+						 &len)) >= 0) {
+					FD_SET(fd, &unixm);
+					dprintf("New UNIX connect assigned to fd: " \
+						"%d.\n", fd);
+					FD_SET(fd, &readfds);
+				} else {
+					dprintf("Error accepting UNIX connection: " \
+						"%d = %s.\n", errno, strerror(errno));
+				}
 			}
-			else {
-				dprintf("Error accepting UNIX connection: " \
-					"%d = %s.\n", errno, strerror(errno));
-			}
-		}
-
 #endif
 
 #ifdef SYSLOG_INET
@@ -1014,6 +1161,24 @@ int main(argc, argv)
 				sleep(10);
 			}
 		}
+#endif
+#else
+		if ( FD_ISSET(fileno(stdin), &readfds) ) {
+			dprintf("Message from stdin.\n");
+			memset(line, '\0', sizeof(line));
+			line[0] = '.';
+			parts[fileno(stdin)] = (char *) 0;
+			i = read(fileno(stdin), line, MAXLINE);
+			if (i > 0) {
+				printchopped(LocalHostName, line, i+1, fileno(stdin));
+		  	} else if (i < 0) {
+		    		if (errno != EINTR) {
+		      			logerror("stdin");
+				}
+		  	}
+			FD_CLR(fileno(stdin), &readfds);
+		  }
+
 #endif
 	}
 }
@@ -1026,21 +1191,27 @@ int usage()
 }
 
 #ifdef SYSLOG_UNIXAF
-static int create_unix_socket()
+static int create_unix_socket(const char *path)
 {
 	struct sockaddr_un sunx;
 	int fd;
 	char line[MAXLINE +1];
 
+	if (path[0] == '\0')
+		return -1;
+
+	(void) unlink(path);
+
+	memset(&sunx, 0, sizeof(sunx));
 	sunx.sun_family = AF_UNIX;
-	(void) strncpy(sunx.sun_path, LogName, sizeof(sunx.sun_path));
+	(void) strncpy(sunx.sun_path, path, sizeof(sunx.sun_path));
 	fd = socket(AF_UNIX, SOCK_STREAM, 0);
 	if (fd < 0 || bind(fd, (struct sockaddr *) &sunx,
 			   sizeof(sunx.sun_family)+strlen(sunx.sun_path)) < 0 ||
-	    chmod(LogName, 0666) < 0 || listen(fd, 5) < 0) {
-		(void) sprintf(line, "cannot create %s", LogName);
+	    chmod(path, 0666) < 0 || listen(fd, 5) < 0) {
+		(void) snprintf(line, sizeof(line), "cannot create %s", path);
 		logerror(line);
-		dprintf("cannot create %s (%d).\n", LogName, errno);
+		dprintf("cannot create %s (%d).\n", path, errno);
 		close(fd);
 #ifndef SYSV
 		die(0);
@@ -1062,9 +1233,9 @@ static int create_inet_socket()
 		return fd;
 	}
 
+	memset(&sin, 0, sizeof(sin));
 	sin.sin_family = AF_INET;
 	sin.sin_port = LogPort;
-	sin.sin_addr.s_addr = 0;
 	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, \
 		       (char *) &on, sizeof(on)) < 0 ) {
 		logerror("setsockopt, suspending inet");
@@ -1116,7 +1287,7 @@ crunch_list(list)
 	 * so we don't have to care about this.
 	 */
 	count = 0;
-	while ((q=index(p, LIST_DELIMITER))) {
+	while ((q=strchr(p, LIST_DELIMITER))) {
 		result[count] = (char *) malloc((q - p + 1) * sizeof(char));
 		if (result[count] == NULL) {
 			printf ("Sorry, can't get enough memory, exiting.\n");
@@ -1204,7 +1375,7 @@ void printchopped(hname, msg, len, fd)
 		{
 			dprintf("Previous: %s\n", tmpline);
 			dprintf("Next: %s\n", msg);
-			strcat(tmpline, msg);
+			strcat(tmpline, msg);	/* length checked above */
 			printline(hname, tmpline);
 			if ( (strlen(msg) + 1) == len )
 				return;
@@ -1250,13 +1421,14 @@ void printline(hname, msg)
 	char *msg;
 {
 	register char *p, *q;
-	register int c;
+	register unsigned char c;
 	char line[MAXLINE + 1];
 	int pri;
 
 	/* test for special codes */
 	pri = DEFUPRI;
 	p = msg;
+
 	if (*p == '<') {
 		pri = 0;
 		while (isdigit(*++p))
@@ -1269,16 +1441,17 @@ void printline(hname, msg)
 	if (pri &~ (LOG_FACMASK|LOG_PRIMASK))
 		pri = DEFUPRI;
 
+	memset (line, 0, sizeof(line));
 	q = line;
-	while ((c = *p++ & 0177) != '\0' &&
-	    q < &line[sizeof(line) - 1])
+	while ((c = *p++) && q < &line[sizeof(line) - 1]) {
 		if (c == '\n')
 			*q++ = ' ';
-		else if (iscntrl(c)) {
+		else if (iscntrl(c)&&(c<0177)) {
 			*q++ = '^';
 			*q++ = c ^ 0100;
 		} else
 			*q++ = c;
+	}
 	*q = '\0';
 
 	logmsg(pri, line, hname, SYNC_FILE);
@@ -1300,7 +1473,7 @@ void printsys(msg)
 	int pri, flags;
 	char *lp;
 
-	(void) sprintf(line, "vmunix: ");
+	(void) snprintf(line, sizeof(line), "vmunix: ");
 	lp = line + strlen(line);
 	for (p = msg; *p != '\0'; ) {
 		flags = ADDDATE;
@@ -1339,8 +1512,7 @@ char *textpri(pri)
 	for (c_fac = facilitynames; c_fac->c_name && !(c_fac->c_val == LOG_FAC(pri)<<3); c_fac++);
 	for (c_pri = prioritynames; c_pri->c_name && !(c_pri->c_val == LOG_PRI(pri)); c_pri++);
 
-	/*	sprintf (res, "%d.%d", LOG_FAC(pri), LOG_PRI(pri));*/
-	sprintf (res, "%s.%s<%d>", c_fac->c_name, c_pri->c_name, pri);
+	snprintf (res, sizeof(res), "%s.%s<%d>", c_fac->c_name, c_pri->c_name, pri);
 
 	return res;
 }
@@ -1402,6 +1574,7 @@ void logmsg(pri, msg, from, flags)
 			untty();
 			fprintlog(f, (char *)from, flags, msg);
 			(void) close(f->f_file);
+			f->f_file = -1;
 		}
 #ifndef SYSV
 		(void) sigsetmask(omask);
@@ -1484,11 +1657,13 @@ void fprintlog(f, from, flags, msg)
 {
 	struct iovec iov[6];
 	register struct iovec *v = iov;
+	char repbuf[80];
+#ifdef SYSLOG_INET
 	register int l;
 	char line[MAXLINE + 1];
-	char repbuf[80];
 	time_t fwd_suspend;
 	struct hostent *hp;
+#endif
 
 	dprintf("Called fprintlog, ");
 
@@ -1508,7 +1683,7 @@ void fprintlog(f, from, flags, msg)
 		v->iov_base = msg;
 		v->iov_len = strlen(msg);
 	} else if (f->f_prevcount > 1) {
-		(void) sprintf(repbuf, "last message repeated %d times",
+		(void) snprintf(repbuf, sizeof(repbuf), "last message repeated %d times",
 		    f->f_prevcount);
 		v->iov_base = repbuf;
 		v->iov_len = strlen(repbuf);
@@ -1526,6 +1701,7 @@ void fprintlog(f, from, flags, msg)
 		dprintf("\n");
 		break;
 
+#ifdef SYSLOG_INET
 	case F_FORW_SUSP:
 		fwd_suspend = time((time_t *) 0) - f->f_time;
 		if ( fwd_suspend >= INET_SUSPEND_TIME ) {
@@ -1567,7 +1743,7 @@ void fprintlog(f, from, flags, msg)
 			}
 			else {
 			        dprintf("%s found, resuming.\n", f->f_un.f_forw.f_hname);
-				bcopy(hp->h_addr, (char *) &f->f_un.f_forw.f_addr.sin_addr, hp->h_length);
+				memcpy((char *) &f->f_un.f_forw.f_addr.sin_addr, hp->h_addr, hp->h_length);
 				f->f_prevcount = 0;
 				f->f_type = F_FORW;
 				goto f_forw;
@@ -1590,9 +1766,8 @@ void fprintlog(f, from, flags, msg)
 			dprintf("Not sending message to remote.\n");
 		else {
 			f->f_time = now;
-			(void) sprintf(line, "<%d>%s", f->f_prevpri, \
+			(void) snprintf(line, sizeof(line), "<%d>%s\n", f->f_prevpri, \
 				(char *) iov[4].iov_base);
-			strcat(line, "\n");	/* ASP */
 			l = strlen(line);
 			if (l > MAXLINE)
 				l = MAXLINE;
@@ -1608,6 +1783,7 @@ void fprintlog(f, from, flags, msg)
 			}
 		}
 		break;
+#endif
 
 	case F_CONSOLE:
 		f->f_time = now;
@@ -1713,20 +1889,18 @@ void wallmsg(f, iov)
 	char p[6 + UNAMESZ];
 	register int i;
 	int ttyf, len;
-	FILE *uf;
 	static int reenter = 0;
 	struct utmp ut;
+	struct utmp *uptr;
 	char greetings[200];
 
 	if (reenter++)
 		return;
 
 	/* open the user login file */
-	if ((uf = fopen(UTMP_FILE, "r")) == NULL) {
-		logerror(UTMP_FILE);
-		reenter = 0;
-		return;
-	}
+	utmpname(_PATH_UTMP);
+	setutent();
+
 
 	/*
 	 * Might as well fork instead of using nonblocking I/O
@@ -1740,13 +1914,14 @@ void wallmsg(f, iov)
 		(void) signal(SIGTTOU, SIG_IGN);
 		(void) sigsetmask(0);
 #endif
-		(void) sprintf(greetings,
+		(void) snprintf(greetings, sizeof(greetings),
 		    "\r\n\7Message from syslogd@%s at %.24s ...\r\n",
 			(char *) iov[2].iov_base, ctime(&now));
 		len = strlen(greetings);
 
 		/* scan the user login file */
-		while (fread((char *) &ut, sizeof(ut), 1, uf) == 1) {
+		while ((uptr = getutent())) {
+			memcpy(&ut, uptr, sizeof(ut));
 			/* is this slot used? */
 			if (ut.ut_name[0] == '\0')
 				continue;
@@ -1798,12 +1973,13 @@ void wallmsg(f, iov)
 		exit(0);
 	}
 	/* close the user login file */
-	(void) fclose(uf);
+	endutent();
 	reenter = 0;
 }
 
 void reapchild()
 {
+	int saved_errno = errno;
 #if defined(SYSV) && !defined(linux)
 	(void) signal(SIGCHLD, reapchild);	/* reset signal handler -ASP */
 	wait ((int *)0);
@@ -1816,6 +1992,7 @@ void reapchild()
 #ifdef linux
 	(void) signal(SIGCHLD, reapchild);	/* reset signal handler -ASP */
 #endif
+	errno = saved_errno;
 }
 
 /*
@@ -1850,7 +2027,7 @@ const char *cvthname(f)
 	 * Notice that the string still contains the fqdn, but your
 	 * hostname and domain are separated by a '\0'.
 	 */
-	if ((p = index(hp->h_name, '.'))) {
+	if ((p = strchr(hp->h_name, '.'))) {
 		if (strcmp(p + 1, LocalDomain) == 0) {
 			*p = '\0';
 			return (hp->h_name);
@@ -1888,6 +2065,7 @@ void domark()
 	int lognum;
 #endif
 
+	if (MarkInterval > 0) {
 	now = time(0);
 	MarkSeq += TIMERINTVL;
 	if (MarkSeq >= MarkInterval) {
@@ -1909,6 +2087,7 @@ void domark()
 			BACKOFF(f);
 		}
 	}
+	}
 	(void) signal(SIGALRM, domark);
 	(void) alarm(TIMERINTVL);
 }
@@ -1933,11 +2112,9 @@ void logerror(type)
 	dprintf("Called logerr, msg: %s\n", type);
 
 	if (errno == 0)
-		(void) sprintf(buf, "syslogd: %s", type);
-	else if ((unsigned) errno > sys_nerr)
-		(void) sprintf(buf, "syslogd: %s: error %d", type, errno);
+		(void) snprintf(buf, sizeof(buf), "syslogd: %s", type);
 	else
-		(void) sprintf(buf, "syslogd: %s: %s", type, sys_errlist[errno]);
+		(void) snprintf(buf, sizeof(buf), "syslogd: %s: %s", type, strerror(errno));
 	errno = 0;
 	logmsg(LOG_SYSLOG|LOG_ERR, buf, LocalHostName, ADDDATE);
 	return;
@@ -1951,6 +2128,11 @@ void die(sig)
 	register struct filed *f;
 	char buf[100];
 	int lognum;
+	int i;
+	int was_initialized = Initialized;
+
+	Initialized = 0;	/* Don't log SIGCHLDs in case we
+				   receive one during exiting */
 
 	for (lognum = 0; lognum <= nlogs; lognum++) {
 		f = &Files[lognum];
@@ -1959,23 +2141,42 @@ void die(sig)
 			fprintlog(f, LocalHostName, 0, (char *)NULL);
 	}
 
+	Initialized = was_initialized;
 	if (sig) {
 		dprintf("syslogd: exiting on signal %d\n", sig);
-		(void) sprintf(buf, "exiting on signal %d", sig);
+		(void) snprintf(buf, sizeof(buf), "exiting on signal %d", sig);
 		errno = 0;
-		logerror(buf);
+		logmsg(LOG_SYSLOG|LOG_INFO, buf, LocalHostName, ADDDATE);
 	}
 
-	/* Close the sockets. */
-        close(funix);
+	/* Close the UNIX sockets. */
+        for (i = 0; i < nfunix; i++)
+		if (funix[i] != -1)
+			close(funix[i]);
+	/* Close the inet socket. */
 	if (InetInuse) close(inetm);
 
 	/* Clean-up files. */
-	(void) unlink(LogName);
+        for (i = 0; i < nfunix; i++)
+		if (funixn[i] && funix[i] != -1)
+			(void)unlink(funixn[i]);
+#ifndef TESTING
 	(void) remove_pid(PidFile);
+#endif
 	exit(0);
 }
 
+/*
+ * Signal handler to terminate the parent process.
+ */
+#ifndef TESTING
+void doexit(sig)
+	int sig;
+{
+	exit (0);
+}
+#endif
+
 /*
  *  INIT -- Initialize syslogd from configuration table
  */
@@ -1984,7 +2185,12 @@ void init()
 {
 	register int i, lognum;
 	register FILE *cf;
-	register struct filed *f, **nextp = (struct filed **) 0;
+	register struct filed *f;
+#ifndef TESTING
+#ifndef SYSV
+	register struct filed **nextp = (struct filed **) 0;
+#endif
+#endif
 	register char *p;
 	register unsigned int Forwarding = 0;
 #ifdef CONT_LINE
@@ -2040,11 +2246,24 @@ void init()
 	}
 	
 
+#ifdef SYSV
+	lognum = 0;
+#else
+	f = NULL;
+#endif
+
 	/* open the configuration file */
 	if ((cf = fopen(ConfFile, "r")) == NULL) {
 		dprintf("cannot open %s.\n", ConfFile);
 #ifdef SYSV
-		cfline("*.ERR\t" _PATH_CONSOLE, *nextp);
+		allocate_log();
+		f = &Files[lognum++];
+#ifndef TESTING
+		cfline("*.err\t" _PATH_CONSOLE, f);
+#else
+		snprintf(cbuf,sizeof(cbuf), "*.*\t%s", ttyname(0));
+		cfline(cbuf, f);
+#endif
 #else
 		*nextp = (struct filed *)calloc(1, sizeof(*f));
 		cfline("*.ERR\t" _PATH_CONSOLE, *nextp);
@@ -2058,11 +2277,6 @@ void init()
 	/*
 	 *  Foreach line in the conf table, open that file.
 	 */
-#ifdef SYSV
-	lognum = 0;
-#else
-	f = NULL;
-#endif
 #if CONT_LINE
 	cline = cbuf;
 	while (fgets(cline, sizeof(cbuf) - (cline - cbuf), cf) != NULL) {
@@ -2079,7 +2293,7 @@ void init()
 #if CONT_LINE
 		strcpy(cline, p);
 #endif
-		for (p = index(cline, '\0'); isspace(*--p););
+		for (p = strchr(cline, '\0'); isspace(*--p););
 #if CONT_LINE
 		if (*p == '\\') {
 			if ((p - cbuf) > BUFSIZ - 30) {
@@ -2115,8 +2329,12 @@ void init()
 	(void) fclose(cf);
 
 #ifdef SYSLOG_UNIXAF
-	if (funix < 0)
-		funix = create_unix_socket();
+	for (i = 0; i < nfunix; i++) {
+		if (funix[i] != -1)
+			close(funix[i]);
+		if ((funix[i] = create_unix_socket(funixn[i])) != -1)
+			dprintf("Opened UNIX socket `%s'.\n", funixn[i]);
+	}
 #endif
 
 #ifdef SYSLOG_INET
@@ -2223,16 +2441,19 @@ void cfline(line, f)
 	int singlpri = 0;
 	int ignorepri = 0;
 	int syncfile;
+#ifdef SYSLOG_INET
 	struct hostent *hp;
+#endif
 	char buf[MAXLINE];
+	char xbuf[200];
 
 	dprintf("cfline(%s)\n", line);
 
-	errno = 0;	/* keep sys_errlist stuff out of logerror messages */
+	errno = 0;	/* keep strerror() stuff out of logerror messages */
 
 	/* clear out file entry */
 #ifndef SYSV
-	bzero((char *) f, sizeof(*f));
+	memset((char *) f, 0, sizeof(*f));
 #endif
 	for (i = 0; i <= LOG_NFACILITIES; i++) {
 		f->f_pmask[i] = TABLE_NOPRI;
@@ -2247,12 +2468,12 @@ void cfline(line, f)
 			continue;
 
 		/* collect priority name */
-		for (bp = buf; *q && !index("\t ,;", *q); )
+		for (bp = buf; *q && !strchr("\t ,;", *q); )
 			*bp++ = *q++;
 		*bp = '\0';
 
 		/* skip cruft */
-		while (index(",;", *q))
+		while (strchr(",;", *q))
 			q++;
 
 		/* decode priority name */
@@ -2276,16 +2497,14 @@ void cfline(line, f)
 		}
 
 		if (pri < 0) {
-			char xbuf[200];
-
-			(void) sprintf(xbuf, "unknown priority name \"%s\"", buf);
+			(void) snprintf(xbuf, sizeof(xbuf), "unknown priority name \"%s\"", buf);
 			logerror(xbuf);
 			return;
 		}
 
 		/* scan facilities */
-		while (*p && !index("\t .;", *p)) {
-			for (bp = buf; *p && !index("\t ,;.", *p); )
+		while (*p && !strchr("\t .;", *p)) {
+			for (bp = buf; *p && !strchr("\t ,;.", *p); )
 				*bp++ = *p++;
 			*bp = '\0';
 			if (*buf == '*') {
@@ -2324,9 +2543,8 @@ void cfline(line, f)
 			} else {
 				i = decode(buf, FacNames);
 				if (i < 0) {
-					char xbuf[200];
 
-					(void) sprintf(xbuf, "unknown facility name \"%s\"", buf);
+					(void) snprintf(xbuf, sizeof(xbuf), "unknown facility name \"%s\"", buf);
 					logerror(xbuf);
 					return;
 				}
@@ -2390,12 +2608,12 @@ void cfline(line, f)
 			f->f_type = F_FORW;
 		}
 
-		bzero((char *) &f->f_un.f_forw.f_addr,
+		memset((char *) &f->f_un.f_forw.f_addr, 0,
 			 sizeof(f->f_un.f_forw.f_addr));
 		f->f_un.f_forw.f_addr.sin_family = AF_INET;
 		f->f_un.f_forw.f_addr.sin_port = LogPort;
 		if ( f->f_type == F_FORW )
-			bcopy(hp->h_addr, (char *) &f->f_un.f_forw.f_addr.sin_addr, hp->h_length);
+			memcpy((char *) &f->f_un.f_forw.f_addr.sin_addr, hp->h_addr, hp->h_length);
 		/*
 		 * Otherwise the host might be unknown due to an
 		 * inaccessible nameserver (perhaps on the same
